# cmd_changelog.py - web interface for bazaar
# this code is derived by the hgweb server of the Matt Mackall's mercurial
# project
#
# Copyright 2006 Goffredo Baroncelli <kreijack@inwind.it>
#
# This file is part of bazaar-webserve.
#
# bazaar-webserve is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# bazaar-webserve is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bazaar-webserve; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from bzrlib.errors import *
from bzrlib.osutils import format_date

import hgweb
from hgweb import (
    WebServeErrorRevisionNotFound,
    WebServeErrorPathNotFound
)


def prev_page(hgweb, revid, **other):

    def prev_page_entry(revids, **other):
        for rev in revids:
                yield hgweb.template_revid(rev, **other)
                yield "<br>"


    revids = hgweb.get_prev_revid(revid, hgweb.maxchanges)
    if len(revids) == 0:
        yield ""
        return

    yield hgweb.t("changelognavigation",
                    title = "Previous page",
                    entries = prev_page_entry(revids, **other),
                )


def next_page(hgweb, revid, **other):

    revids = hgweb.revision_history.get_parents(revid)
    if len(revids) == 0:
        yield ""
        return

    yield hgweb.t("changelognavigation",
                    title = "Next page",
                    entries = hgweb.template_revid(revids[0], **other),
                )


def get_heads(hgweb, revid, **other):

    heads = hgweb.revision_history.get_branch_head(revid)

    def get_entries(**other):
        for rev in heads:
                if rev == revid:
                    continue
                yield hgweb.template_revid(rev, **other)
                yield "<br>"

    if len(heads) == 0:
        return ""

    if len(heads) == 1 and heads[0] == revid:
        return ""

    return hgweb.t("changelognavigation",
                    title = "Heads",
                    entries = get_entries(**other),
                )


def changelog(hgweb, revid, search=None, path=None,
    pathrevid=None, otherrevid=None ):

    def where_merged(revid, **other):

        m_revs = hgweb.revision_history.get_merge_point_list(revid)
        m_revs = hgweb.revision_history.simplify_merge_point_list(m_revs)
        if len(m_revs) == 0:
            yield ""
            return

        def where_merged_entry(m_revs, **other):
            for rev in m_revs:
                yield hgweb.template_revid(rev, **other)
                yield "<br>"

        yield hgweb.t("changelognavigation",
                    title="Merged point",
                    entries=where_merged_entry(m_revs, **other),
                )

    def changelist(cuts_revid, **other):
        parity = 0
        l = []
        cuts_revisions = hgweb.branch.repository.get_revisions(cuts_revid)
        for rev in cuts_revisions:
            rev_id = rev.revision_id
            try:
                date_str = format_date(rev.timestamp,
                           rev.timezone or 0,
                           'original')

                def get_parents(**map):
                    if len(rev.parent_ids) > 1:
                        for parent in rev.parent_ids:
                            yield hgweb.t("changelogparent",
                              revlink =hgweb.template_revid(parent, **other)
                            )

                dodiff = ""
                if otherrevid and otherrevid <> rev.revision_id:
                    dodiff = hgweb.t("dodiff", otherrevid=otherrevid,
                        revid=rev.revision_id, path=path and path or "",
                        pathrevid=path and pathrevid or "")

                r = hgweb.t(
                    'changelogentry',
                    parity = parity,
                    #rev = flag and revno or '['+str(revno)+']',
                    node = rev.revision_id,
                    author = rev.committer,
                    desc = rev.message,
                    date = rev.timestamp,
                    parent = get_parents,
                    dodiff = dodiff,
                    otherrevid = rev.revision_id,
                    revid = revid,
                    path = path and path or "",
                    pathrevid = pathrevid,
                    revlink = hgweb.template_revid(rev.revision_id, **other)
                )
            except (KeyError, NoSuchRevision):
                # in case of missing parent
                r = hgweb.t(
                    'changelogentry',
                    parity = parity,
                    rev = "missing",
                    node = rev_id,
                    author = "missing",
                    desc = "missing",
                    date = 0,
                    revlink = rev_id,
                )

            parity = 1 - parity

            yield r

    revid, history, pathrevid, otherrevid = hgweb.compute_history(revid, path,
                                                    pathrevid, otherrevid )

    if search != None:
        lsearch = search.lower()
        count = len(history)

        spos=0
        while spos < count:

            rev_id = history[spos]

            if rev_id.lower() == lsearch:
                break
            spos = spos + 1

        #otherwise search in the  message
        if spos>=count:
            spos=0
            while spos < count:

                rev_id = history[spos]

                try:
                    rev = hgweb.branch.repository.get_revision(rev_id)
                except (KeyError, NoSuchRevision):
                    raise WebServeErrorRevisionNotFound(rev_id)

                if rev.message.lower().find(lsearch) >= 0:
                    break
                spos = spos + 1

        if spos < count:
            pos = spos
        else:
            yield hgweb.t("textnotfound", search=search)
            return

        history = history[pos:]

    searchfrom = history[0]
    if search != None:
        searchfrom = history[1]
    cut_revs = history[:hgweb.maxchanges]

    if path:
        changelogpath = hgweb.t("changelogpath",
            pathlink = hgweb.template_path(revid, path, pathrevid))

        def nav(**map):
            return hgweb.template_nav_revid("cflnaventry",
                        revid, path = path, pathrevid = pathrevid,
                        pagesize=hgweb.maxchanges,
                        otherrevid=otherrevid)
    else:
        changelogpath = ""

        def nav(**map):
            return hgweb.template_nav_revid("naventry",
                        revid, pagesize=hgweb.maxchanges,
                        otherrevid=otherrevid)

    if otherrevid:
        diffrevision = hgweb.t("diffrevision",
                           revid=revid,
                           path = path and path or "",
                           pathrevid = pathrevid,
                           link = hgweb.template_revid(otherrevid,
                                path=path or "",
                                otherrevid=otherrevid or "",
                                pathrevid=pathrevid or ""))
    else:
        diffrevision = ""

    (pos, revid, merge_depth, revno, end_of_merge) = \
        hgweb.revision_history.get_revision_info(revid)

    def prevpage(**map):
        return prev_page(hgweb, cut_revs[0],
                                path=path or "",
                                otherrevid=otherrevid or "",
                                pathrevid=pathrevid or "")

    def nextpage(**map):
        return next_page(hgweb, cut_revs[-1],
                                path=path or "",
                                otherrevid=otherrevid or "",
                                pathrevid=pathrevid or "")

    yield hgweb.t('changelog',
                 changenav = nav,
                 rev = pos,
                 revid = revid,
                 path = path and path or "",
                 pathrevid = pathrevid and pathrevid or "",
                 searchfrom = searchfrom,
                 otherrevid = otherrevid,
                 changesets = 10,
                 mergepoints = where_merged(revid,
                                path=path or "",
                                otherrevid=otherrevid or "",
                                pathrevid=pathrevid or ""),
                 changelogpath = changelogpath,
                 diffrevision = diffrevision,
                 entries = changelist(cut_revs,
                                path=path or "",
                                otherrevid=otherrevid or "",
                                pathrevid=pathrevid or ""),
                prevpage = prevpage,
                nextpage = nextpage,
                heads = get_heads(hgweb, cut_revs[0],
                                path=path or "",
                                otherrevid=otherrevid or "",
                                pathrevid=pathrevid or ""),

            )

