# hgweb.py - web interface for bazaar
#
# Copyright 21 May 2005 - (c) 2005 Jake Edge <jake@edge2.net>
# Copyright 2005 Matt Mackall <mpm@selenic.com>
#
# This file is part of bazaar-webserve.
#
# bazaar-webserve is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# bazaar-webserve is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bazaar-webserve; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import cgi
import time
import re
import difflib
import socket
import sys
import zlib
import urllib
import exceptions
from cStringIO import StringIO

import bzrlib
import bzrlib.branch
import bzrlib.tsort
from bzrlib.osutils import format_date
from bzrlib.diff import show_diff_trees
from bzrlib.errors import *
from bzrlib.revisionspec import RevisionSpec
from bzrlib.annotate import _annotate_file
from bzrlib.errors import *
from bzrlib.inventory import ROOT_ID
from bzrlib.errors import BzrError

from templater import (
    templater,
    common_filters,
    write2,
)


def templatepath():
    p = os.path.join(os.getcwd(), "templates")
    if os.path.isdir(p):
        return p

    for f in "templates", "../templates":
        p = os.path.join(os.path.dirname(__file__), f)
        if os.path.isdir(p):
            return p


class WebserveRevisionList:

    def __init__(self, branch):
        self._branch = branch
        self._load_data()

    def _revno_list_revno(self, revno):
        return ".".join([str(n) for n in revno])

    def _load_data(self):

        self._short_history = self._branch.revision_history()
        self._last_revid = self._short_history[-1]
        self._revision_graph = self._branch.repository.get_revision_graph()
        self._merge_sort = bzrlib.tsort.merge_sort(self._revision_graph,
                                        self._last_revid, generate_revno=True)

        self._full_history = list()
        self._revision_info = dict()
        self._revno_revid = dict()
        count=0
        for (seq, revid, merge_depth, revno, end_of_merge) in self._merge_sort:
            self._full_history.append(revid)
            revno_str = self._revno_list_revno(revno)
            self._revno_revid[revno_str] = revid
            self._revision_info[revid] = (seq, revid, merge_depth, revno_str,
                                            end_of_merge)
            count = count +1
        self._count = count
        self._where_merged = dict()

        for revid in self._revision_graph.keys():
            if not revid in self._full_history:
                continue
            for parent in self._revision_graph[revid]:
                if self._where_merged.has_key(parent):
                    if not revid in self._where_merged[parent]:
                        self._where_merged[parent].append(revid)
                else:
                    self._where_merged[parent] = [revid,]

    def _get_full_revision_history_by_fileid(self, file_id):

        w = self._branch.repository.weave_store.get_weave(file_id,
                self._branch.repository.get_transaction())

        _revids = w.versions()
        h = self._full_history
        revids = []

        for i in h:
            if i in _revids:
                revids.append(i)

        return revids

    def get_short_revision_history(self):
        return self._short_history

    def get_full_revision_history(self):
        return self._full_history

    def get_short_revision_history_from(self, revid):
        """return the short revision_history, starting from revid"""
        while True:
            yield revid
            if not self._revision_graph.has_key(revid):
                return
            parents = self._revision_graph[revid]
            if len(parents) == 0:
                return
            revid = self._revision_graph[revid][0]

    def get_short_revision_history_by_fileid_from(self, revid, file_id):
        rev_file_id_list = self._get_full_revision_history_by_fileid(file_id)
        while True:
            if revid in rev_file_id_list:
                yield revid
            if not self._revision_graph.has_key(revid):
                return
            parents = self._revision_graph[revid]
            if len(parents) == 0:
                return
            revid = parents[0]

    def get_last_revid(self):
        return self._last_revid

    def get_parents(self, revid):
        return self._revision_graph[revid]

    def get_merged(self, revid):
        return self._where_merged[revid]

    def get_revision_info(self, revid):
        return self._revision_info[revid]

    def revno2revid(self, revno):
        return self._revno_revid[revno]

    def exist_revision(self, revision):
        return revision in self._full_history

    def get_count(self):
        return self._count

    def get_seq(self, revid):
        (seq, revid, merge_depth, revno_str, end_of_merge) = \
            self._revision_info[revid]
        return seq

    def get_where_merged(self, revid):
        try:
            return self._where_merged[revid]
        except:
            return []

    def revid2revno(self, revid):
        (seq, revid, merge_depth, revno, end_of_merge) = \
            self.get_revision_info(revid)
        return revno

    def get_merge_point_list__(self, revid):
        """Return the list where the node is merged"""

        if revid in self._short_history:
            return []

        merge_point = []
        while True:

            children = self.get_where_merged(revid)
            nexts = []
            for child in children:
                child_parents = self.get_parents(child)
                if child_parents[0] == revid:
                    nexts.append(child)
                else:
                    merge_point.append(child)

            if len(nexts) == 0:
                # only merge
                return merge_point

            while len(nexts) > 1:
                # branch
                next = nexts.pop()
                merge_point_next = self.get_merge_point_list__(next)
                merge_point.extend(merge_point_next)

            revid = nexts[0]

    def simplify_merge_point_list(self, revids):
        """if a revision is already merged, don't show further merge points"""
        d = {}
        for revid in revids:
            revno = self.revid2revno(revid)
            revnol = revno.split(".")
            revnos = ".".join(revnol[:-2])
            revnolast = int(revnol[-1])
            if d.has_key(revnos):
                m = d[revnos][0]
                if revnolast < m:
                    d[revnos] = (revnolast, revid)
            else:
                d[revnos] = (revnolast, revid)

        return [d[revnos][1] for revnos in d.keys()]

    def get_branch_head_and_merge_point(self, revid):
        """Return the list where the node is merged"""

        if revid in self._short_history:
            return [self._short_history[-1]], []

        merge_point = []
        heads = []
        while True:

            children = self.get_where_merged(revid)
            nexts = []
            for child in children:
                child_parents = self.get_parents(child)
                if child_parents[0] == revid:
                    nexts.append(child)
                else:
                    merge_point.append(child)

            if len(nexts) == 0:
                heads.append(revid)
                return heads, merge_point

            while len(nexts) > 1:
                # branch
                next = nexts.pop()
                head_next, mpoint_next = self.get_branch_head(next)
                heads.extend(head_next)
                merge_point.extend(mpoint_next)

            revid = nexts[0]

    def get_merge_point_list(self, revid):
        r1 = self.get_branch_head_and_merge_point(revid)[1]
        r2 = self.get_merge_point_list__(revid)
        if(r1 <> r2):
            raise "Errore !!!!"
        return r1

    def get_branch_head(self, revid):
        heads = self.get_branch_head_and_merge_point(revid)[0]
        return heads


class WebServeError(BzrError):

    _fmt = "%(_error)s"

    def __init(self, msg):
        self._error = msg


class WebServeErrorRevisionNotFound(BzrError):

    _fmt = "The revision %(_revid)s doesn't exist"

    def __init__(self, revid):
        self._revid = revid


class WebServeErrorPathNotFound(BzrError):

    _fmt = "The path '%(_path)s'@%(_pathrevid) doesn't exist at revision %(_revid)s"

    def __init__(self, revid, path, pathrevid):
        self._revid = revid
        self._path = path
        self._pathrevid = pathrevid


class WebServeErrorGeneric(BzrError):

    _fmt = "%(_error)s"

    def __init(self, htmlmsg):
        self._htmlmsg = htmlmsg
        self._error = re.sub("<[^>]*>", "", htmlmsg)


class hgweb:

    def __init__(self, path, name=None, templates="", mapfile="map"):
        self.templates = templates
        self.reponame = name
        self.path = path
        self.stream = sys.stdout
        self.mtime = -1
        self.viewonly = 0
        self.refreshed = 0
        self.mapfile = mapfile
        self.commentfilter = []

        # set to 1 if you want the lock enabled
        self.dolock = 1

        # Set to 1 if you want the profile info at the bottom of the web page
        self.profile = 0

        # set to 1 to allow a download of a tar.gz archive
        self.tararchive = 0

        # number of revision in changelog page
        self.maxchanges = 20

        self.has_tags = False
        self.encoding = 'utf_8'

    def filter_comment(self, txt, flt):
        """Filter the string s"""
        if not flt:
            return txt

        # FIXME: find a better separator
        for f in flt:
            if f[:8] == "replace:":
                import re
                f = f[8:]
                (s, r) = f.split(f[0])[1:3]
                txt = re.sub(s, r, txt)
            else:
                raise "unknow filter"
        return txt

    def refresh(self):
        if self.path.startswith("http://") or \
           self.path.startswith("https://") or \
           self.path.startswith("sftp://") or \
           self.path.startswith("ftp://") or \
           self.path.startswith("sftp://"):
            # refresh only 1st time
            if self.refreshed:
                return
            self.refreshed = 1
        else:
            for p in [os.path.join(self.path, ".bzr", "revision-history"),
                os.path.join(self.path, ".bzr", "branch", "revision-history")]:
                    if not os.access(p, os.R_OK):
                        continue
                    s = os.stat(p)
                    if s.st_mtime == self.mtime:
                        return
                    self.mtime = s.st_mtime

        self.branch = bzrlib.branch.Branch.open(self.path)
        self.revision_history = WebserveRevisionList(self.branch)

    def date(self, cs):
        return time.asctime(time.gmtime(float(cs[2].split(' ')[0])))

    def get_fileid(self, path, revid):
        inv = self.branch.repository.get_revision_inventory(revid)
        return inv.path2id(path)

    def test_exist_file(self, rev, path, pathrevid):

        file_id= self.get_fileid(path, pathrevid)

        if file_id == None:
            return False

        inv = self.branch.repository.get_revision_inventory(rev)
        return inv.has_id(file_id)

    def get_revid(self, revid, default=False):
        if not revid and default != False:
            return default
        if not revid:
            return self.revision_history.get_last_revid()
        for i in revid:
            if not i in "0123456789.":
                return revid
        try:
            return self.revision_history.revno2revid(revid)
        except:
            return None

    def template_revid(self, rev, **other):
        (seq, revid, merge_depth, revno, end_of_merge) = \
            self.revision_history.get_revision_info(rev)
        yield self.t("tmplrevid", revno=revno, revid=rev, **other)

    def template_path(self, rev, path, pathrevid=None):

        if pathrevid == None:
            short = 1
            pathrevid = rev
        else:
            short = 0

        if path == '/':
            path =""
        fpath = ""
        lpath = path.split('/')

        yield self.t("roottpdir", rev = rev, fullpath=fpath,
            pathrevid=pathrevid, elem="")

        if path != "":
            for e in lpath[:-1]:
                fpath+=e+'/'
                yield self.t("tpdir", rev = rev, fullpath=fpath,
                    pathrevid=pathrevid, elem=e)

        if short == 0:
            yield self.t("tpfname", rev = rev, fullpath=path,
                pathrevid=pathrevid, elem=lpath[-1],
                atpathrevid=self.template_revid(pathrevid))
        else:
            yield self.t("tpfnameshort", rev = rev, fullpath=path,
                pathrevid=pathrevid, elem=lpath[-1],
                atpathrevid=self.template_revid(pathrevid))

    def template_nav(self, template, pos, history, pagesize=1, **other):
        return self.template_nav_revid(template, history[pos-1][1],
                            pagesize, **other)

    def template_nav_revid(self, template, revid, pagesize=1, **other):

        count = self.revision_history.get_count()
        pos = self.revision_history.get_seq(revid)
        history = self.revision_history.get_full_revision_history()

        def navigation(template, revid, **other):
            # TODO: why we have to recalculate these values ??
            count = self.revision_history.get_count()
            pos = self.revision_history.get_seq(revid)
            history = self.revision_history.get_full_revision_history()

            def seq(factor = 1):

                yield 1 * factor
                yield 3 * factor

                for f in seq(factor * 10):
                    yield f

            l = []
            res = ""
            for f in seq():
                if f > count:
                    break
                r = "%d" % f

                if pos + f < count:
                    l.insert(0, ("-" + r, pos + f))
                if pos - f >= 0:
                    l.append(("+" + r, pos - f))

            #yield self.t( "navigationpos", pos=pos, count=count )

            yield self.t(template, revid = history[-1], label="(1)",
                **other)

            for label, pos in l:
                yield self.t(template, label = label,
                    revid = history[pos],
                    **other)

            yield self.t(template, label="tip", **other)

        if pos >0:
            xpos = pos - pagesize -1
            if xpos <1:
                xpos = 1
            next = self.t(template, revid = history[xpos-1], label="&gt;",
                **other)
        else:
            next = "&gt;"

        if pos < count:
            xpos = pos + pagesize -1
            if xpos >= count:
                xpos = count

            prev = self.t(template, revid = history[xpos-1], label="&lt;",
                **other)
        else:
            prev = "&lt;"

        yield self.t("navigation", pagenext=next, pageprev=prev,
            pos = count-pos, count = count,
            navigation = navigation(template, revid, **other))

    def list_dir_revs_v6(self, file_id):
        h=set()
        history = self.revision_history.get_full_revision_history()
        w = self.branch.repository.get_inventory_weave()
        #for lineno, insert, deleteset, line in w.walk():
        for line, rev in w.iter_lines_added_or_present_in_versions(
                set(history)):
            start = line.find('file_id="')+9
            if start < 9:
                continue
            end = line.find('"', start)
            assert end >= 0
            fid = line[start:end]

            start = line.find('revision="')+10
            if start < 10:
                continue
            end = line.find('"', start)
            assert end>= 0
            revid = line[start:end]

            if fid == file_id:
                h.add(revid)
                continue

            start = line.find('parent_id="')+11
            if start < 11:
                if file_id == ROOT_ID:
                    h.add(revid)
                continue

            end = line.find('"', start)
            assert end>= 0
            parent_id = line[start:end]
            if parent_id == file_id:
                h.add(revid)
                continue

        return h

    def compute_history(self, revid, path=None, pathrevid=None, otherrevid=None):

        revid = self.get_revid(revid)
        if not self.revision_history.exist_revision(revid):
            raise WebServeErrorRevisionNotFound(revid)

        otherrevid = self.get_revid(otherrevid, "")
        if otherrevid and not self.revision_history.exist_revision(otherrevid):
            raise WebServeErrorRevisionNotFound(otherrevid)
            return

        if path <> None:

            pathrevid = self.get_revid(pathrevid, revid)

            if not self.revision_history.exist_revision(pathrevid):
                raise WebServeErrorRevisionNotFound(pathrevid)

            file_id = self.get_fileid(path, pathrevid)
            inv = self.branch.repository.get_revision_inventory(revid)
            if file_id == None or not inv.has_id(file_id):
                raise WebServeErrorPathNotFound(revid, path, pathrevid)

            if path[-1] == '/':
                h = self.list_dir_revs_v6(file_id)
                history = []
                i=0
                for rid in self.revision_history.get_full_revision_history():
                    if rid in h:
                        history.append(rid)
            else:
                history = list(
                    self.revision_history.get_short_revision_history_by_fileid_from(revid, file_id)
                )

            if not revid in history:
                revid = inv[file_id].revision

        else:

            history = list(self.revision_history.get_short_revision_history_from(revid))
            path=""
            pathrevid=""

        return revid, history, pathrevid, otherrevid

    def get_prev_revid(self, revid_from, n):
        res = []
        revid = revid_from
        while n>0:
            children = self.revision_history.get_where_merged(revid)
            nexts = []
            for child in children:
                parents = self.revision_history.get_parents(child)
                if parents[0] == revid:
                    nexts.append(child)

            if len(nexts) == 0:
                break

            for next in nexts[1:]:
                res.extend(self.get_prev_revid(next, n-1))

            revid = nexts[0]
            n = n - 1

        res.append(revid)

        return [x  for x in res if x<>revid_from]

    #
    # Command implementation
    #

    def changelog(self, revid, search=None, path=None,
        pathrevid=None, otherrevid=None):
            import cmd_changelog
            return cmd_changelog.changelog(self, revid, search, path,
                pathrevid, otherrevid)

    def content(self, revid, path, pathrevid, annotate=False):
            import cmd_content
            return cmd_content.content(self, revid, path, pathrevid, annotate)

    def dump_content(self, revid, path, pathrevid):
            import cmd_content
            return cmd_content.dump_content(self, revid, path, pathrevid)

    def diff2revs(self, revid, otherrevid=None, path=None, pathrevid=None):

        new_tree = self.branch.repository.revision_tree(revid)
        old_tree = self.branch.repository.revision_tree(otherrevid)

        if path:
            fileid=self.get_fileid(path, pathrevid)
            pathlist=[new_tree.id2path(fileid)]
        else:
            pathlist=None

        # FIXME:the path is not pathrevid related
        delta = new_tree.changes_from(old_tree, specific_files=pathlist)

        def diff(**map):

            def dodiff(b, old_tree, new_tree):
                s = StringIO()
                show_diff_trees(old_tree, new_tree, s, specific_files=pathlist,
                    old_label="", new_label="")
                return s.getvalue().split("\n")

            def prettyprintlines(diff):
                first = True
                for l in diff:
                    if not first:
                        yield("\n")
                    else:
                        first = False
                    if l.startswith('='):
                        yield self.t("difflineinfo",
                            line=l.decode('utf_8', 'replace'))
                    elif l.startswith('+++'):
                        yield self.t("difflineplus",
                            line=l.decode('utf_8', 'replace'))
                    elif l.startswith('---'):
                        yield self.t("difflineminus",
                            line=l.decode('utf_8', 'replace'))
                    elif l.startswith('+'):
                        yield self.t("difflineplus",
                            line=l.decode(self.encoding, 'replace'))
                    elif l.startswith('-'):
                        yield self.t("difflineminus",
                            line=l.decode(self.encoding, 'replace'))
                    elif l.startswith('@'):
                        yield self.t("difflineat", line=l)
                    else:
                        yield self.t("diffline",
                                     line=l.decode(self.encoding, 'replace'))

            yield self.t("diffblock",
                         lines=prettyprintlines(dodiff(self.branch,
                                                       old_tree, new_tree)),
                         parity=0)

        def files(files):
            for path, fid, kind in files:
                if kind == 'directory':
                    path += '/'
                elif kind == 'symlink':
                    path += '@'
                yield path+'<br>\n'

        def files_renamed(files):
            for oldpath, newpath, fid, kind, text_modified, meta_modified in files:
                if kind == 'directory':
                    yield self.t("revrename", oldpath = '/'+oldpath,
                        newpath = self.template_path(revid, newpath+'/'))
                elif kind == 'symlink':
                    yield self.t("revrename", oldpath = '/'+oldpath,
                        newpath = self.t("revlinkentry",
                        str = self.template_path(revid, newpath)))
                else:
                    yield self.t("revrename", oldpath = '/'+oldpath,
                        newpath = self.template_path(revid, newpath))

        def fileentry(files):
            for path, fid, kind in files:
                if kind == 'directory':
                    yield self.template_path(revid, path+'/')
                elif kind == 'symlink':
                    yield self.t("revlinkentry",
                        str = self.template_path(revid, path))
                else:
                    yield self.template_path(revid, path)

        def fileremoved(files):
            for path, fid, kind in files:
                if kind == 'directory':
                    yield self.t("revdelete", path=path+'/')
                elif kind == 'symlink':
                    yield self.t("revdelete", path='@'+path)
                else:
                    yield self.t("revdelete", path=path)

        yield self.t('diffentry',
            diff = diff,
            filesadded = lambda **map: fileentry(delta.added),
            filesmodified = lambda **map: fileentry([(path,id,kind)
                        for path, id, kind, text_modified, meta_modified in
                            delta.modified]),
            filesrenamed = lambda **map: files_renamed(delta.renamed),
            filesremoved = lambda **map: fileremoved(delta.removed),
        )

    def diffpage(self, revid, otherrevid=None, path=None, pathrevid=None):

        revid, history, pathrevid, otherrevid = \
            self.compute_history(revid, path, pathrevid, otherrevid)

        # compute the delta
        rev = self.branch.repository.get_revision(revid)
        if not otherrevid and len(rev.parent_ids):
            otherrevid = rev.parent_ids[0]

        def getparents(rev):
            for pid in rev.parent_ids:
                yield self.template_revid(pid)
                yield "<br>"

        if path:
            if not self.test_exist_file(otherrevid, path, pathrevid):
                yield self.t("pathnotfound", path=path,
                    pathrevid=pathrevid, rev=otherrevid)
                return
            if not self.test_exist_file(revid, path, pathrevid):
                yield self.t("pathnotfound", path=path,
                    pathrevid=pathrevid, rev=revid)
                return

            changelogpath = self.t("changelogpath",
                pathlink = self.template_path(revid, path, pathrevid))
        else:
            changelogpath = ""

        yield self.t('diffpage',
            diff = self.diff2revs(revid, otherrevid, path, pathrevid),
            linkrevid = self.template_revid(revid),
            linkotherrevid = self.template_revid(otherrevid),
            diffnav = lambda **map: self.template_nav_revid("diffnaventry",
                                        revid, otherrevid=otherrevid,
                                        path=path and path or "",
                                        pathrevid=path and pathrevid or ""),
            changelogpath = changelogpath,
            revid = rev.revision_id,
            otherrevid = otherrevid,
            path = path and path or "",
            pathrevid = path and pathrevid or ""
        )

    def revision(self, revid):

        revid = self.get_revid(revid)
        if not self.revision_history.exist_revision(revid):
            raise WebServeErrorRevisionNotFound(revid)

        # compute the delta
        rev = self.branch.repository.get_revision(revid)
        if len(rev.parent_ids):
            previous = rev.parent_ids[0]
        else:
            previous = None

        def getparents(rev):
            for pid in rev.parent_ids:
                yield self.template_revid(pid)
                yield "<br>"

        def getchildren(rev):
            for cid in self.revision_history.get_where_merged(rev):
                yield self.template_revid(cid)
                yield "<br>"

        yield self.t('revision',
            diff = self.diff2revs(revid, previous),
            revisionnav = lambda **map:
                    self.template_nav_revid("revnaventry", revid),
            parents = getparents(rev),
            children = getchildren(revid),
            linkrev = self.template_revid(rev.revision_id),
            node = rev.revision_id,
            author = rev.committer,
            desc = rev.message,
            date = rev.timestamp,
        )

    def inventory(self, rev, path=None, pathrevid=None):
        import cmd_inventory
        return cmd_inventory.inventory(self, rev, path, pathrevid)

    def showtags(self):

        def list_tags():
            for tag_name, target in self.branch.tags.get_tag_dict().items():
                yield self.t("tagitem",
                    id=self.template_revid(target),
                    name=tag_name)

        return self.t("tagspage", tagslist=list_tags())

    # TODO unify with the code of export command
    def export_archive(self, name, revid, path="", mode=""):

        revid = self.get_revid(revid)

        def tar_exporter(tree, dest, root, filter_func=None, compression=None):
            """Export this tree to a new tar file.

            `dest` will be created holding the contents of this tree; if it
            already exists, it will be clobbered, like with "tar -c".
            """
            from time import time
            import tarfile
            now = time()
            compression = str(compression or '')
            if root is None:
                root = get_root_name(dest)
            try:
                if isinstance(dest, basestring):
                    ball = tarfile.open(dest, 'w:' + compression)
                else:
                    ball = tarfile.open("none", 'w:' + compression, dest)
            except tarfile.CompressionError, e:
                raise BzrError(str(e))
            #mutter('export version %r' % tree)
            inv = tree.inventory
            for dp, ie in inv.iter_entries():
                if filter_func and filter_func(dp, ie):
                    continue
                #mutter("  export {%s} kind %s to %s" % (ie.file_id, ie.kind, dest))
                item, fileobj = ie.get_tar_item(root, dp, now, tree)
                ball.addfile(item, fileobj)
            ball.close()

        def zip_exporter(tree, dest, root, filter_func=None):
            """export this tree to a zip file"""

            from time import time
            import zipfile
            now = time()

            if root is None:
                root = get_root_name(dest)

            try:
                ball = zipfile.ZipFile(dest, 'w', zipfile.ZIP_DEFLATED)
            except zipfile.BadZipfile, e:
                raise BzrError(str(e))

            inv = tree.inventory
            for dp, ie in inv.iter_entries():
                if filter_func and filter_func(dp, ie):
                    continue
                # FIXME: we have to handle also directories and symlinks
                if ie.kind <> "file": continue
                item, fileobj = ie.get_tar_item(root, dp, now, tree)
                ball.writestr(str(item.name), fileobj.read())
            ball.close()

        tree = self.branch.repository.revision_tree(revid)
        root = name + "_" + revid

        if path:
            f = lambda dp, ie : not dp.startswith(path)
        else:
            f = None

        # FIXME: it is right to use application/force-download
        # see http://lists.nyphp.org/pipermail/talk/2003-September/005058.html

        #
        # header("Cache-Control: no-cache, must-revalidate");
        # header("Pragma: no-cache");
        # header("Content-type: application/octet-stream");
        # header("Content-Disposition: attachment; filename=\"$filename\"");
        #

        if mode == "tgz":
            dest = root + ".tar.gz"
            self.raw_write('Content-Type: application/force-download\n'  \
              'Content-Disposition: attachment; filename=%s\n\n'%(dest))
            tar_exporter(tree, sys.stream, root, filter_func=f, compression="gz")
        elif mode == "tbz2":
            dest = root + ".tar.bz2"
            self.raw_write('Content-Type: application/force-download\n'  \
              'Content-Disposition: attachment; filename=%s\n\n'%(dest))
            tar_exporter(tree, sys.stream, root, filter_func=f, compression="bz2")
        elif mode == "tar":
            dest = root + ".tar"
            self.raw_write('Content-Type: application/force-download\n'  \
              'Content-Disposition: attachment; filename=%s\n\n'%(dest))
            tar_exporter(tree, self.stream, root, filter_func=f, compression="")
        else: # mode == zip
            dest = root + ".zip"
            fo = os.tmpfile()
            self.raw_write('Content-Type: application/force-download\n'  \
                'Content-Disposition: attachment; filename=%s\n\n'%(dest))
            zip_exporter(tree, fo, root, filter_func=f)
            fo.seek(0)
            self.raw_write(fo.read())

    def raw_write(self, *args):
        for arg in args:
            self.stream.write(arg)

    def _write(self, *arg):
        if not self.profile:
            return write2(self.stream, *arg)

        try:
            tmpfilename = os.tempnam()
            import hotshot
            import hotshot.stats

            def w(*arg):
                return write2(self.stream, *arg)

            prof = hotshot.Profile(tmpfilename)
            prof.runcall(w, *arg)
            prof.close()

            #sys.stdout.write(w.get_stream( ).getvalue( ))
            write2(self.stream, """
                <hr width=100%>
                <pre>
            """)

            stats = hotshot.stats.load(tmpfilename)
            stats.strip_dirs()
            stats.sort_stats('time', 'file', 'calls')
            tmp = sys.stdout
            sys.stdout = self.stream
            stats.print_stats()
            sys.stdout = tmp
        except:
            try:
                os.unlink(tmpfilename)
            except:
                pass
            raise

        os.unlink(tmpfilename)
        write2(self.stream, """
            </pre>
            <hr width=100%>
        """)

        write2(self.t("footer"))
        return

    def write(self, *arg):
        try:
            self._write(*arg)
        except WebServeErrorRevisionNotFound, e:
            revid = e._revid
            return self.write(self.t("revisionnotfound", rev=revid))
        except WebServeErrorPathNotFound, e:
            revid = e._revid
            path = e._path
            pathrevid = e._pathrevid
            return self.write(self.t("pathnotfound", path=path,
                    pathrevid = pathrevid, rev = revid))

    def do_cmd(self, url, args):

        def header(**map):
            yield self.t("header", **map)

        def footer(**map):
            if not self.profile:
                yield self.t("footer", **map)

        def get(arg):
            if args.has_key(arg):
                return args[arg][0]
            else:
                return None

        def filter_comment__(s):
            return self.filter_comment(s, self.commentfilter)

        def tagsbutton_function(**map):
            if self.has_tags:
                return self.t("tagsbutton")
            else:
                return ""

        # try to decode from utf8
        try:
            _url = url.decode('utf-8', 'replace')
            _args = {}
            for k in args.keys( ):
                if k in ["revid", "pathrevid", "otherrevid", "rev"]:
                    _args[k] = args[k]
                else:
                    _k = k.decode('utf-8', 'replace')
                    _args[_k] = [i.decode('utf-8', 'replace') for i in args[k]]

            url = _url
            args = _args
        except exceptions.UnicodeDecodeError:
            pass

        self.refresh()

        last_revid = self.revision_history.get_last_revid()

        t = self.templates or templatepath()
        m = os.path.join(t, self.mapfile)
        if args.has_key('style'):
            b = os.path.basename("map-" + args['style'][0])
            p = os.path.join(t, b)
            if os.path.isfile(p):
                m = p

        name = self.reponame
        self.has_tags = (hasattr(self.branch, "tags") and
            self.branch.tags.supports_tags())

        self.t = templater(m, common_filters,
                           {"url":        url,
                            "repo":       name,
                            "header":     header,
                            "footer":     footer,
                            "newline":    "\n",
                            "tagsbutton": tagsbutton_function,
                            "documentencoding": "UTF-8",
                            })
        self.t.add_filters("commentfilter", filter_comment__)

        if not args.has_key('cmd'):
            args['cmd'] = [self.t.cache['default'], ]

        if args.has_key('rev'):
            revno = args['rev'][0]
        else:
            revno = last_revid

        if self.dolock:
            self.branch.repository.lock_read()
        try:


            if args['cmd'][0] == 'changelog':

                otherrevid = get("otherrevid")
                pathrevid = get("pathrevid")
                path = get("path")

                self.write(self.changelog(revno, None, path, pathrevid, otherrevid))

            elif args['cmd'][0] == 'search':
                w = get('w')
                otherrevid = get("otherrevid")
                pathrevid = get("pathrevid")
                path = get("path")
                self.write(self.changelog(revno, search=w,
                    path=path, pathrevid=pathrevid, otherrevid=otherrevid))

            elif (args['cmd'][0] == 'tags' and self.has_tags):
                self.write(self.showtags())

            elif args['cmd'][0] == 'revision':
                if args.has_key("revid"):
                    self.write(self.revision(args['revid'][0]))
                else:
                    self.write(self.revision(str(revno)))

            elif args['cmd'][0] == 'diff':
                revid = get("rev")
                otherrevid = get("otherrevid")
                path = get("path")
                pathrevid = get("pathrevid")
                self.write(self.diffpage(revid, otherrevid, path, pathrevid))

            elif args['cmd'][0] == 'inventory':
                path = pathrevid = None

                if args.has_key('path'):
                    path = args['path'][0]

                if args.has_key('pathrevid'):
                    pathrevid = args['pathrevid'][0]

                self.write(self.inventory(revno, path, pathrevid))

            elif args['cmd'][0] == 'content':

                path = pathrevid = None
                revid = last_revid

                if args.has_key('rev'):
                    revid = args['rev'][0]

                if args.has_key('pathrevid'):
                    pathrevid = args['pathrevid'][0]

                if args.has_key('path'):
                    path = args['path'][0]

                self.write(self.content(revid, path, pathrevid))

            elif args['cmd'][0] == 'dump':

                path = pathrevid = None
                revid = last_revid

                if args.has_key('rev'):
                    revid = args['rev'][0]

                if args.has_key('pathrevid'):
                    pathrevid = args['pathrevid'][0]

                if args.has_key('path'):
                    path = args['path'][0]

                self.dump_content(revid, path, pathrevid)


            elif args['cmd'][0] == 'annotate':

                path = pathrevid = None
                revid = last_revid

                if args.has_key('rev'):
                    revid = args['rev'][0]

                if args.has_key('pathrevid'):
                    pathrevid = args['pathrevid'][0]

                if args.has_key('path'):
                    path = args['path'][0]

                self.write(self.content(revid, path, pathrevid, annotate=True))


            elif self.tararchive and args['cmd'][0] == 'export':
                path = get("path")
                mode = get("mode")
                self.export_archive(name, revno, path and path or "", mode)

            else:
                self.write(self.t("error", command=args['cmd'][0]))

        finally:
                if self.dolock:
                    self.branch.repository.unlock()

    def run(self, stream=None):
        port = os.environ["SERVER_PORT"]
        port = port != "80" and (":" + port) or ""
        uri = os.environ["REQUEST_URI"]
        if "?" in uri:
            uri = uri.split("?")[0]
        url = "http://%s%s%s" % (os.environ["SERVER_NAME"], port, uri)
        args = cgi.parse()
        if(stream):
            self.stream = stream
        self.do_cmd(url, args)


def create_server(path, name, templates, address, port, use_ipv6,
                  accesslog, errorlog, dolock, tararchive,
                  profile, mapfile, option_provide_file,
                  commentfilter, encoding):

    def openlog(opt, default):
        if opt and opt != '-':
            return open(opt, 'w')
        return default

    accesslog = sys.stdout
    errorlog = sys.stderr

    import BaseHTTPServer

    class IPv6HTTPServer(BaseHTTPServer.HTTPServer):
        address_family = getattr(socket, 'AF_INET6', None)

        def __init__(self, *args, **kwargs):
            if self.address_family is None:
                raise RepoError('IPv6 not available on this system')
            BaseHTTPServer.HTTPServer.__init__(self, *args, **kwargs)

    class hgwebhandler(BaseHTTPServer.BaseHTTPRequestHandler):

        def provide_file(self, root, path):
            path = urllib.unquote(path)
            if path.find("..")>=0:
                self.send_error(404, "Uncorrect path [path=%s]\n"%path)
                #sys.stderr.write("Uncorrect path [path=%s]\n"%path)
                return

            newpath = os.path.join(root, path[1:])
            if not os.access(newpath, os.R_OK):
                self.send_error(404, "File isn't accessible [path=%s]\n"%newpath)
                #sys.stderr.write("Uncorrect path [path=%s]\n"%path)
                return

            self.send_response(200, "Script output follows")
            sys.stdout.write('Content-Type: bzr/data\n'  \
                    'Content-Disposition: attachment; filename=%s\n\n'%(
                    path.encode("utf_8","replace")))
            #sys.stderr.write("path=%s\n"%os.path.join(self.path,path))
            if os.environ['REQUEST_METHOD'] != "HEAD":
                fp = open(newpath)
                sys.stdout.write(fp.read())
                fp.close()

        def log_error(self, format, *args):
            errorlog.write("%s - - [%s] %s\n"%(self.address_string(),
                                                 self.log_date_time_string(),
                                                 format % args))

        def log_message(self, format, *args):
            accesslog.write("%s - - [%s] %s\n"%(self.address_string(),
                                                  self.log_date_time_string(),
                                                  format % args))

        def do_HEAD(self):
            self.do_POST()

        def do_POST(self):
            try:
                self.do_hgweb()
            except socket.error, inst:
                if inst.args[0] != 32:
                    raise

        def do_GET(self):
            self.do_POST()

        def do_hgweb(self):

            env = {}
            env['GATEWAY_INTERFACE'] = 'CGI/1.1'
            env['REQUEST_METHOD'] = self.command
            env['SERVER_NAME'] = self.server.server_name
            env['SERVER_PORT'] = str(self.server.server_port)
            env['REQUEST_URI'] = "/"

            query = ""
            p = self.path.find("?")
            if p>=0:
                query = self.path[p + 1:]
                query = query.replace('+', ' ')
            else:
                # ugly hack !!!!!!
                p=self.path.find(".bzr/")
                if(p >= 0 and option_provide_file):
                    os.environ.update(env)
                    sys.stdin = self.rfile
                    sys.stdout = self.wfile
                    self.provide_file(hg.path, self.path)

                    return

            if query:
                env['QUERY_STRING'] = query
            host = self.address_string()
            if host != self.client_address[0]:
                env['REMOTE_HOST'] = host
                env['REMOTE_ADDR'] = self.client_address[0]

            if self.headers.typeheader is None:
                env['CONTENT_TYPE'] = self.headers.type
            else:
                env['CONTENT_TYPE'] = self.headers.typeheader
            length = self.headers.getheader('content-length')
            if length:
                env['CONTENT_LENGTH'] = length
            accept = []
            for line in self.headers.getallmatchingheaders('accept'):
                if line[:1] in "\t\n\r ":
                    accept.append(line.strip())
                else:
                    accept = accept + line[7:].split(',')
            env['HTTP_ACCEPT'] = ','.join(accept)

            os.environ.update(env)

            #save = sys.argv, sys.stdin, sys.stdout, sys.stderr
            #try:
                #sys.stdin = self.rfile
                #sys.stdout = self.wfile
                #sys.argv = ["hgweb.py"]
            if '=' not in query:
                sys.argv.append(query)
            self.send_response(200, "Script output follows")
            hg.run(self.wfile)
            #finally:
                #sys.argv, sys.stdin, sys.stdout, sys.stderr = save

    hg = hgweb(path, name, templates, mapfile)
    if dolock:
        hg.dolock=1
    if tararchive:
        hg.tararchive=1
    if profile:
        hg.profile=1
    if commentfilter:
        hg.commentfilter = commentfilter
    if not encoding is None:
        hg.encoding = encoding

    if use_ipv6:
        return IPv6HTTPServer((address, port), hgwebhandler)
    else:
        return BaseHTTPServer.HTTPServer((address, port), hgwebhandler)


class hgwebdir:

    class _cfg_klass:
        """wrapper to configobj"""

        def __init__(self):
            self._cfg = None

        def read(self, filename):
            """read the filename"""
            from bzrlib.util.configobj.configobj import ConfigObj

            l = [i for i in open(filename).readlines() if not i.rstrip().startswith(";")]

            self._cfg = ConfigObj(l)

        def sections(self):
            return [x for x in self._cfg.keys() if x<> "DEFAULT"]

        def get(self, section, key, default_value=None):
            if not section in self._cfg.keys():
                raise "section not found"
            if not key in self._cfg.get(section).keys():
                if not default_value is None:
                    return default_value
                if section <> "DEFAULT":
                    return self.get("DEFAULT", key)

                raise "key not found"

            return  self._cfg.get(section).get(key)

    def __init__(self, config):
        self.cp = self._cfg_klass()
        self.cp.read(config)
        self.stream = sys.stdout

    def run(self, stream=None):

        def header(**map):
            yield tmpl("header", **map)

        def footer(**map):
            yield tmpl("footer", **map)

        def get(sec, val, default):
            # TODO: replace with the cp.get( )
            try:
                return self.cp.get(sec, val).decode("utf-8")
            except:
                return default

        if(stream):
            self.stream = stream
        templates = get("DEFAULT", "template", "")
        if not templates:
            templates = templatepath()

        style = get("DEFAULT", "style", "")
        if style == "":
            mapfile = "map"
        else:
            mapfile = "map-"+style
        m = os.path.join(templates, mapfile)

        tmpl = templater(m, common_filters,
                         {"header": header, "footer": footer})

        try:
            virtual = os.environ["PATH_INFO"].strip('/')
        except:
            virtual = ""

        if virtual:

            if not virtual in self.cp.sections():
                write2(self.stream, tmpl("reponotfound", repository = virtual))
                return

            real = self.cp.get(virtual, "path")

            style = get(virtual, "style", "")
            if style == "":
                mapfile = "map"
            else:
                mapfile = "map-"+style

            h = hgweb(real, get(virtual, "name", virtual),
                get(virtual, "template", ""), mapfile)

            h.dolock = int(get(virtual, "lock", "0"))
            h.profile = int(get(virtual, "profile", "0"))
            h.tararchive = int(get(virtual, "tararchive", "0"))
            commentfilter = get(virtual, "commentfilter", None)
            if(commentfilter):
                h.commentfilter = commentfilter.split(";;")
            h.commentfilter.extend(get(virtual, "commentfilterlist", []))
            h.encoding = get(virtual, "encoding",
                os.environ.get('BZR_ENCODING', 'utf_8'))

            h.run()
            return

        def show_project(project, parity):
            r = self.cp.get(project, "path")

            url = '/'+os.environ["REQUEST_URI"]+'/'+project
            while url.find("//") >= 0:
                url = url.replace("//", "/")

            if r.startswith("http://") or \
               r.startswith("https://") or \
               r.startswith("sftp://") or \
               r.startswith("ftp://") or \
               r.startswith("sftp://"):
                lastupdate = 0
                repo = r
            else:
                lastupdate=0
                for p in [os.path.join(r, ".bzr", "revision-history"),
                     os.path.join(r, ".bzr", "branch","revision-history")]:
                    if not os.access(p, os.R_OK):
                        continue
                    lastupdate = os.stat(p).st_mtime
                repo = url

            repo = get(project, "repo", repo)

            return tmpl("indexentry",
                       author = get(project, "author", "unknown"),
                       name = get(project, "name", project),
                       email = get(project, "email", ""),
                       repo = repo,
                       url = url,
                       parity = parity,
                       shortdesc = get(project, "description", "unknown"),
                       lastupdate = lastupdate
            )

        def entries():
            sects = self.cp.sections()
            parity = 0
            if not "MAIN" in sects:

                yield tmpl("index_table_header")
                l = self.cp.sections()
                l.sort()
                for v in l:
                    yield show_project(v, parity)
                    parity = 1 - parity
                yield tmpl("index_table_footer")
                return

            opened_table = False
            groups = self.cp.get("MAIN", "groups_list")
            for group in groups:

                repos = self.cp.get(group, "repositories")
                title = self.cp.get(group, "title", group)
                description = self.cp.get(group, "description", "")

                yield tmpl("index_project_header",
                    title=title,
                    description=description)

                parity = 0
                yield tmpl("index_table_header")
                for repo in repos:
                    yield show_project(repo, parity)
                    parity = 1 - parity
                yield tmpl("index_table_footer")

        write2(self.stream, tmpl("index", entries = entries()))


def cgi_start_webserve(path, title, templates):
    import cgitb
    cgitb.enable()

    h = hgweb(path, title, templates)
    h.run()


def cgi_start_webserve_dir(config):
    import cgitb
    cgitb.enable()

    h = hgwebdir(config)
    h.run()

