# test_webserve.py - test suite for the bazaar web interface
#
# Copyright 2006 Goffredo Baroncelli <kreijack@inwind.it>
#
# This file is part of bazaar-webserve.
#
# bazaar-webserve is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# bazaar-webserve is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bazaar-webserve; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import sys
import StringIO
import os
import re
import zipfile

import bzrlib
from bzrlib.tests.blackbox import ExternalBase
from bzrlib.tests import TestCaseInTempDir
from bzrlib.workingtree import WorkingTree

import hgweb
from templater import (
        templater,
        write2,
)


class _TestWebserveBase(ExternalBase):

    def do_test(self, url, repo=None):
        """Call the webserve engine, with the url "url"
           If repo==Null, the repo is in the current working dir
        """

        def url2args(url):
            d = dict()
            if len(url)<1:
                return d
            if url[0].find("?")>=0:
                url = url[url.find("?")+1:]

            for i in url.split(";"):
                key, value = i.split("=")
                d[key]=(value,)

            return d

        if repo is None:
            repo = os.getcwd()

        s = StringIO.StringIO()
        oldstdout = sys.stdout
        sys.stdout = s
        hg = hgweb.hgweb(repo, "TEST")

        self.update_hgweb(hg)

        hg.do_cmd(url, url2args(url))
        sys.stdout = oldstdout

        return s.getvalue()

    def update_hgweb(self, hg):
        pass

    def do_test_text(self, url, repo=None):
        s = self.do_test(url, repo)

        start = s.find("<")
        while start >=0:
            end = s.find(">", start)
            if end < 0: break
            s = s[:start] + s[end+1:]
            start = s.find("<")

        s = re.sub("[ \t]+", " ", s)
        while s.find("\n \n")>=0:
            s = s.replace("\n \n", "\n")
        return s

    def assertContains(self, haystack, needle_re):
        """Assert that a contains something matching a regular expression."""
        if haystack.find(needle_re)<0:
            raise AssertionError('string "%s" not found in "%s"'
                    % (needle_re, haystack))

    def assertContainsReMultiline(self, haystack, needle_re):
        """Assert that a contains something matching a regular expression."""
        if not re.search(needle_re, haystack, re.DOTALL):
            raise AssertionError('pattern "%s" not found in "%s"'
                    % (needle_re, haystack))

    def assertNotContainsReMultiline(self, haystack, needle_re):
        """Assert that a does not match a regular expression"""
        if re.search(needle_re, haystack, re.DOTALL):
            raise AssertionError('pattern "%s" found in "%s"'
                    % (needle_re, haystack))

    def appendFile(self, path, content):
        file = open(path, "ab")
        file.write(content)
        file.close()


class TestWebserve(_TestWebserveBase):

    def _prepare(self):

        wt = self.make_branch_and_tree('.')
        self.build_tree_contents([
            ('hello.txt', "hello_content\n"),
            ('goodbye.txt', "goodbye_content\n"),
            ('meep_dir/',),
            ('meep_dir/meep.txt', "meep_content\n")])
        wt.add(["hello.txt"])
        wt.commit("commit1")
        wt.add(["goodbye.txt"])
        wt.commit("commit2")
        wt.add(["meep_dir"])
        wt.add(["meep_dir/meep.txt"])
        wt.remove("hello.txt")
        os.symlink("meep_dir/meep.txt", "link")
        wt.add(["link"])
        wt.commit("commit3")
        wt.rename_one("meep_dir/meep.txt", "./meep_moved.txt")
        wt.commit("commit4")

    def test_log_simple(self):

        self._prepare()

        log = self.do_test("")
        self.assertContainsRe(log, "commit1")
        self.assertContainsRe(log, "commit2")
        self.assertContainsRe(log, "commit3")

        log = self.do_test_text("?cmd=changelog;rev=1")
        self.assertContainsRe(log, "commit1")

    def test_log_wrong_revision(self):

        self._prepare()

        log = self.do_test_text("?cmd=changelog;rev=x7x")
        self.assertContainsReMultiline(log, "Warning.*"
            "The revision.*"
            "x7x.*"
            "doesn't exist")

    def test_revision_simple(self):

        self._prepare()

        log = self.do_test_text("?cmd=revision;rev=1")
        self.assertContainsRe(log, "commit1")
        self.assertContainsRe(log, "hello.txt")
        self.assertContainsRe(log, "\+hello_content")

        log = self.do_test_text("?cmd=revision;revid=3")
        self.assertContainsRe(log, "commit3")
        self.assertContainsRe(log, "hello.txt")
        self.assertContainsRe(log, "meep.txt")
        self.assertContainsRe(log, "-hello_content")
        self.assertContainsRe(log, "\+meep_content")

        log = self.do_test("?cmd=revision;rev=x7x")
        self.assertContainsReMultiline(log, "Warning.*"
            "The revision.*"
            "x7x.*"
            "doesn't exist")

    def test_inventory_simple(self):

        self._prepare()

        log = self.do_test("?cmd=inventory;rev=1")
        self.assertContainsRe(log, "hello.txt")
        log = self.do_test("?cmd=inventory;rev=2")
        self.assertContainsRe(log, "hello.txt")
        self.assertContainsRe(log, "goodbye.txt")
        log = self.do_test("?cmd=inventory;rev=3")
        self.assertContainsRe(log, "goodbye.txt")
        self.assertContainsRe(log, "meep_dir/")
        self.assertContainsRe(log, "@.*link")
        log = self.do_test("?cmd=inventory;rev=3;path=meep_dir/")
        self.assertContainsRe(log, "meep.txt")

    def test_inventory_wrong_revid_path(self):

        self._prepare()

        log = self.do_test("?cmd=inventory;rev=x7x")
        self.assertContainsReMultiline(log, "Warning.*"
            "The revision.*"
            "x7x.*"
            "doesn't exist")

        log = self.do_test("?cmd=inventory;rev=1;path=foo")
        self.assertContainsReMultiline(log, "Warning.*"
            "The path.*"
            "foo.*"
            "doesn't exist at revision")

        log = self.do_test("?cmd=inventory;rev=3;path=meep_dir/;pathrevid=x7x")
        self.assertContainsReMultiline(log, "Warning.*"
            "The revision.*"
            "x7x.*"
            "doesn't exist")

    def test_content_simple(self):

        self._prepare()

        log = self.do_test("?cmd=content;rev=3;path=meep_dir/meep.txt")
        self.assertContainsRe(log, "meep_content")
        self.assertContainsRe(log, "commit3")

        log = self.do_test("?cmd=content;rev=3;path=link")
        self.assertContainsReMultiline(log, "link.*-&gt;.*meep_dir/meep.txt")

    def test_content_wrong_path_wrong_revid(self):

        self._prepare()

        log = self.do_test("?cmd=content;path=link;rev=x7x")
        self.assertContainsReMultiline(log, "Warning.*"
            "The revision.*"
            "x7x.*"
            "doesn't exist")

        log = self.do_test("?cmd=content;path=wrong_link;rev=3")
        self.assertContainsReMultiline(log, "Warning.*"
            "The path.*"
            "wrong_link.*"
            "doesn't exist at revision")

        log = self.do_test("?cmd=content;rev=3;path=link;pathrevid=x7x")
        self.assertContainsReMultiline(log, "Warning.*"
            "The revision.*"
            "x7x.*"
            "doesn't exist")

    def test_moved_content(self):

        self._prepare()

        log = self.do_test("?cmd=content;rev=3;path=meep_moved.txt;pathrevid=4")
        self.assertContainsRe(log, "meep_content")
        self.assertContainsRe(log, "commit3")

        log = self.do_test("?cmd=content;rev=4;path=meep_dir/meep.txt;pathrevid=3")
        self.assertContainsRe(log, "meep_content")
        self.assertContainsRe(log, "commit4")

    def test_diff_simple(self):

        self._prepare()

        log = self.do_test("?cmd=diff;rev=3;otherrevid=1")
        self.assertContainsRe(log, "-hello_content")
        self.assertContainsRe(log, "\+meep_content")
        self.assertContainsRe(log, "\+goodbye_content")

    def test_diff_raw(self):

        self._prepare()

        log = self.do_test("?cmd=diff;rev=3;otherrevid=1;style=raw")
        log_raw = log[log.find("\n\n")+2:]
        log_cmd, err = self.run_bzr("diff -r 1..3", retcode=1)
        self.assertEqual(log_raw, log_cmd)

    def test_diff_wrong_revid_wrong_path(self):

        self._prepare()

        log = self.do_test_text("?cmd=diff;rev=x7x;otherrevid=1")
        self.assertContainsReMultiline(log, "Warning.*"
            "The revision.*"
            "x7x.*"
            "doesn't exist")

        log = self.do_test_text("?cmd=diff;rev=3;otherrevid=x7x")
        self.assertContainsReMultiline(log, "Warning.*"
            "The revision.*"
            "x7x.*"
            "doesn't exist")

        log = self.do_test_text("?cmd=diff;rev=3;path=/;pathrevid=x7x;otherrevid=1")
        self.assertContainsReMultiline(log, "Warning.*"
            "The revision.*"
            "x7x.*"
            "doesn't exist")

        log = self.do_test_text("?cmd=diff;rev=3;path=/foo;pathrevid=3;otherrevid=1")
        self.assertContainsReMultiline(log, "Warning.*"
            "The path.*"
            "foo.*"
            "doesn't exist at revision")

    def test_annotate(self):

        wt = self.make_branch_and_tree('.')
        self.build_tree_contents([
            ('goodbye.txt', "goodbye_content\n"),])
        wt.add(["goodbye.txt"])
        wt.commit("goodbye_commit1")
        self.appendFile("goodbye.txt", "goodbye_new_line\n")
        wt.commit("goodbye_commit_new_line")
        self.appendFile("goodbye.txt", "goodbye_another_new_line\n")
        wt.commit("goodbye_commit_another_new_line")

        log = self.do_test_text("?cmd=annotate;rev=3;path=goodbye.txt")

        self.assertContainsReMultiline(log, "1:.*goodbye_content.*"
                                    "2:.*goodbye_new_line.*"
                                    "3:.*goodbye_another_new_line")

        self.assertContainsReMultiline(log, "Description:.*goodbye_commit_another_new_line")

        log = self.do_test_text("?cmd=annotate;rev=2;path=goodbye.txt")

        self.assertContainsRe(log, "1:.*goodbye_content")
        self.assertContainsRe(log, "2:.*goodbye_new_line")
        self.assertNotContainsRe(log, "3:.*goodbye_another_new_line")
        self.assertContainsReMultiline(log, "Description:.*goodbye_commit_new_line")
        self.assertNotContainsReMultiline(log, "Description:.*goodbye_commit_another_new_line")

    def test_annotate_wrong_path_wrong_revid(self):

        self._prepare()

        log = self.do_test("?cmd=annotate;path=link;rev=x7x")
        self.assertContainsReMultiline(log, "Warning.*"
            "The revision.*"
            "x7x.*"
            "doesn't exist")

        log = self.do_test("?cmd=annotate;path=wrong_link;rev=3")
        self.assertContainsReMultiline(log, "Warning.*"
            "The path.*"
            "wrong_link.*"
            "doesn't exist at revision")

        log = self.do_test("?cmd=annotate;rev=3;path=link;pathrevid=x7x")
        self.assertContainsReMultiline(log, "Warning.*"
            "The revision.*"
            "x7x.*"
            "doesn't exist")

    def test_merge(self):

        os.mkdir("A")
        os.chdir("A")
        wtA = self.make_branch_and_tree('.')
        wtA.commit("goodbye_commit1") # revno: 1

        for i in range(3):
            wtA.commit("goodbye_commit_%04d"%(i+2)) # revno: 2..4

        os.chdir("..")
        wtB = wtA.bzrdir.clone("B").open_workingtree()
        os.chdir("B")
        for i in range(3):
            wtB.commit("goodbye_commit_branch_B_%04d"%i)

        os.chdir("../A")
        wtA.merge_from_branch(wtB.branch)
        wtA.commit("goodbye_branch_B_merged") # revno: 5

        log = self.do_test_text("?cmd=changelog;rev=5")
        self.assertContainsReMultiline(log,
            "goodbye_branch_B_merged.*"       # description
            "Revision:.*5:.*"               # revision
            "Parent:.*4:.*"
            "Parent:.*4.1.3.*:.*"

            "goodbye_commit_0004.*"           # description
            "Revision:.*4:.*"               # revision

            "goodbye_commit_0003.*"           # description
            "Revision:.*3:.*"               # revision

            "goodbye_commit_0002.*"           # description
            "Revision:.*2:.*"               # revision

            "goodbye_commit1.*"               # description
            "Revision:.*1:"                 # revision
        )

        self.assertNotContainsRe(log, "Heads")

        log = self.do_test_text("?cmd=changelog;rev=4.1.2")
        self.assertContainsReMultiline(log,
            "Heads.*4.1.3:.*"                   # heads

            "Merged point.*5:.*"                # where merge

            "Previous page.*4.1.3.*:.*"         # previous page

            "goodbye_commit_branch_B_0001.*"    # description
            "Revision:.*4.1.2.*:.*"             # revision

            "goodbye_commit_branch_B_0000.*"    # description
            "Revision:.*4.1.1.*:.*"             # revision

            "goodbye_commit_0004.*"             # description
            "Revision:.*4:.*"                   # revision
        )

        self.assertNotContainsRe(log, "goodbye_commit_branch_B_0002")

    def test_content_raw(self):

        wt = self.make_branch_and_tree('.')
        self.build_tree_contents([
            ('goodbye.txt', "goodbye\ncontent\n"),])
        wt.add(["goodbye.txt"])
        wt.commit("goodbye_commit1") # revno 1
        self.appendFile("goodbye.txt", "new_line_no_newline")
        wt.commit("goodbye_commit2") # revno 2
        self.appendFile("goodbye.txt", "\nno ascii code '\xff'\n")
        wt.commit("goodbye_commit3") # revno 2


        log = self.do_test("?cmd=dump;rev=1;path=goodbye.txt")
        raw = log[log.find("\n\n")+2:]
        self.assertEqual(raw, "goodbye\ncontent\n")

        log = self.do_test("?cmd=dump;rev=2;path=goodbye.txt")
        raw = log[log.find("\n\n")+2:]
        self.assertEqual(raw, "goodbye\ncontent\nnew_line_no_newline")

        #FIXME webserve can't handle non ascii char !!!!!
        log = self.do_test("?cmd=dump;rev=3;path=goodbye.txt")
        raw = log[log.find("\n\n")+2:]
        self.assertEqual(raw, "goodbye\ncontent\nnew_line_no_newline\nno ascii code '\xff'\n")


    def test_log_file(self):

        wt = self.make_branch_and_tree('.')
        self.build_tree_contents([
            ('goodbye.txt', "goodbye_content\n"),
            ('hello.txt', "hello_content\n")])
        wt.add(["goodbye.txt", "hello.txt"])
        wt.commit("commit_1") # revno: 1

        self.appendFile("goodbye.txt", "line\n")
        wt.commit("commit_goodbye_2") # revno: 2

        self.appendFile("hello.txt", "line\n")
        wt.commit("commit_hello_3") # revno: 3

        self.appendFile("goodbye.txt", "line\n")
        wt.commit("commit_goodbye_4") # revno: 4

        self.appendFile("hello.txt", "line\n")
        wt.commit("commit_hello_5") # revno: 5

        self.appendFile("hello.txt", "line\n")
        wt.commit("commit_hello_6") # revno: 6

        log = self.do_test("?cmd=changelog;rev=6;path=goodbye.txt")
        self.assertNotContainsRe(log, "commit_hello")
        self.assertContainsRe(log, "commit_goodbye_2")
        self.assertContainsRe(log, "commit_goodbye_4")

        log = self.do_test("?cmd=changelog;rev=6;path=hello.txt")
        self.assertContainsRe(log, "commit_hello")
        self.assertNotContainsRe(log, "commit_goodbye")

    def test_log_dir(self):

        wt = self.make_branch_and_tree('.')
        self.build_tree_contents([
            ('goodbye_dir/',),
            ('goodbye_dir/goodbye.txt', "goodbye_content\n"),
            ('hello.txt', "hello_content\n")])
        wt.add(["goodbye_dir", "goodbye_dir/goodbye.txt", "hello.txt"])
        wt.commit("commit_1") # revno: 1
        self.appendFile("goodbye_dir/goodbye.txt", "line\n")
        wt.commit("commit_goodbye_2") # revno: 2
        self.appendFile("hello.txt", "line\n")
        wt.commit("commit_hello_3") # revno: 3
        self.appendFile("goodbye_dir/goodbye.txt", "line\n")
        wt.commit("commit_goodbye_4") # revno: 4
        self.appendFile("hello.txt", "line\n")
        wt.commit("commit_hello_5") # revno: 5
        self.appendFile("hello.txt", "line\n")
        wt.commit("commit_hello_6") # revno: 6

        log = self.do_test("?cmd=changelog;rev=6;path=goodbye_dir/")
        self.assertNotContainsRe(log, "commit_hello")
        self.assertContainsReMultiline(log, "commit_goodbye_4.*"
                                            "commit_goodbye_2")

        log = self.do_test("?cmd=changelog;rev=6;path=hello.txt")
        self.assertContainsRe(log, "commit_hello")
        self.assertNotContainsRe(log, "commit_goodbye")

    def test_rss(self):

        self._prepare()

        log = self.do_test("?cmd=changelog;style=rss")
        self.assertContainsRe(log, "commit1")
        self.assertContainsRe(log, "commit2")
        self.assertContainsRe(log, "commit3")


class TestWebserveFilterComment(_TestWebserveBase):

    def update_hgweb(self, hg):
        hg.commentfilter=["replace:/text1/text2/", "replace:/text3/text4/"]

    def test_comment_filter(self):
        wt = self.make_branch_and_tree('.')
        self.build_tree_contents([("hello.txt", "goodbye_content\n")])
        wt.add(["hello.txt"])
        wt.commit("text1_text3_text5") # revno: 1
        log = self.do_test_text("?cmd=changelog")
        self.assertNotContainsRe(log, "text[13]")
        self.assertContainsRe(log, "text[245]")


class TestWebserveSearch(_TestWebserveBase):

    def update_hgweb(self, hg):
        hg.maxchanges=5

    def test_search(self):

        wt = self.make_branch_and_tree('.')
        wt.commit("goodbye_commit1")

        for i in range(10):
            wt.commit("goodbye_commit_%04d"%i)
        wt.commit("goodbye_commit_yes_001")
        wt.commit("goodbye_commit_no_002")
        wt.commit("goodbye_commit_yes_003")

        log = self.do_test_text("?cmd=search;w=goodbye_commit_0007")
        self.assertContainsRe(log, "goodbye_commit_0007")
        self.assertContainsRe(log, "goodbye_commit_0006")
        self.assertNotContainsRe(log, "goodbye_commit_0008")

        log = self.do_test("?cmd=search;w=goodbye_commit_0007")

    def test_search_from(self):
        wt = self.make_branch_and_tree('.')
        self.build_tree_contents([
            ('goodbye.txt', "goodbye_content\n"),])
        wt.add(["goodbye.txt"])
        wt.commit("goodbye_commit1")        #1

        self.appendFile("goodbye.txt", "line\n")
        wt.commit("goodbye_commit_yes_001") #2
        self.appendFile("goodbye.txt", "line\n")
        wt.commit("goodbye_commit_no_002")  #3
        self.appendFile("goodbye.txt", "line\n")
        wt.commit("goodbye_commit_yes_003") #4

        log = self.do_test_text("?cmd=search;rev=4;w=goodbye_commit_no")
        self.assertNotContainsRe(log, "goodbye_commit_yes_003")
        self.assertContainsRe(log, "goodbye_commit_yes_001")
        self.assertContainsRe(log, "goodbye_commit_no_002")

        log = self.do_test_text("?cmd=search;rev=3; w=goodbye_commit_yes")
        self.assertNotContainsRe(log, "goodbye_commit_yes_003")
        self.assertContainsRe(log, "goodbye_commit_yes_001")
        self.assertContainsRe(log, "goodbye_commit_no_002")


class TestWebserveTag(_TestWebserveBase):

    # as of 0.14, the default format doesn't do tags so we need to use a
    # specific format
    def make_branch_and_tree(self, relpath):
        import bzrlib.bzrdir
        from bzrlib.tests import TestCaseWithTransport

        format = bzrlib.bzrdir.format_registry.make_bzrdir('dirstate-with-subtree')
        return TestCaseWithTransport.make_branch_and_tree(self, relpath,
            format=format)

    def test_tag_page(self):
        wt = self.make_branch_and_tree('.')
        for i in range(10, 20):
            wt.commit("commit_numebr_%d"%i)
            out, err = self.run_bzr('tag -d . TAG_%d'%i)

        log = self.do_test_text("?cmd=tags")
        j = 0
        for i in range(10, 20):
            j += 1
            self.assertContainsRe(log, 'TAG_%d.*%d:'%(i, j))


class TestWebservePages(_TestWebserveBase):

    def update_hgweb(self, hg):
        hg.maxchanges=5

    def test_page_prev(self):

        wt = self.make_branch_and_tree('.')
        wt.commit("goodbye_commit1")

        for i in range(10):
            wt.commit("goodbye_commit_%04d"%i)

        log = self.do_test_text("?cmd=changelog;rev=3")
        self.assertContainsReMultiline(log, "Previous page.*8:.*"
                                   "goodbye_commit_0001.*"
                                   "Revision:.*3:.*"
        )

    def test_page_next(self):

        wt = self.make_branch_and_tree('.')
        wt.commit("goodbye_commit1")

        for i in range(8):
            wt.commit("goodbye_commit_%04d"%i)

        log = self.do_test_text("?cmd=changelog;rev=8")
        self.assertContainsReMultiline(log, "Previous page.*9:.*"

                                   "goodbye_commit_0006.*"
                                   "Revision:.*8.*"

                                   "goodbye_commit_0002.*"
                                   "Revision:.*4.*"

                                   "Next page.*3:.*"

        )


class TestWebserveArchiver(_TestWebserveBase):

    def update_hgweb(self, hg):
        hg.tararchive = 1

    def test_zip(self):

        wt = self.make_branch_and_tree('.')
        self.build_tree_contents([
            ('goodbye.txt', "goodbye\ncontent\n"),
            ('goodbye2.txt', "goodbye1\ncontent\n"),
            ('goodbye3/',),
            ('goodbye3/goodbye.txt', "goodbye/goodbye\ncontent\n"),
            ('goodbye.bin', "\xff")
        ])
        wt.add(["goodbye.txt", 'goodbye2.txt','goodbye3/',
            'goodbye.bin', 'goodbye3/goodbye.txt' ])
        wt.commit("goodbye_commit") # revno 1


        log = self.do_test("?cmd=export;rev=1;path=goodbye.txt;mode=zip;path=")
        raw = log[log.find("\n\n")+2:]

        #open("/tmp/log.zip","wb").write(raw)

        ios = StringIO.StringIO(raw)
        z = zipfile.ZipFile(ios,"r")

        for fn in z.namelist( ):
            # check the goodbye word  in the name
            self.assertContainsRe(fn, "goodbye")

            if fn.endswith(".txt"):
                # check the goodbye word in the content
                self.assertContainsRe(z.read(fn), "goodbye")

            if fn.endswith("goodbye.bin"):
                self.assertContainsRe(z.read(fn), "\xff")

class TestWebserveEncoding(_TestWebserveBase):

    alpha = u'\u03b1'
    omega = u'\u03c9'

    def update_hgweb(self, hg):
        hg.encoding = self.encoding

    def _test_encoding(self, encoding, text_raw, text_utf8):

        self.encoding = encoding

        wt = self.make_branch_and_tree('.')
        self.build_tree_contents([ ('data.txt', text_raw) ])
        wt.add(["data.txt" ])
        wt.commit("commit")

        log = self.do_test("?cmd=content;rev=1;path=data.txt")

        self.assertContains(log, text_utf8)
        self.assertContains(log, "charset=UTF-8")

    def test_utf8(self):
        self._test_encoding("utf_8",
            "\xc3\xa0\xc3\xa8\xc3\xa9\xc3\xb9\xc3\xac\xc3\xb2",
            "\xc3\xa0\xc3\xa8\xc3\xa9\xc3\xb9\xc3\xac\xc3\xb2"
        )

    def test_iso8859_15(self):
        self._test_encoding("iso8859_15",
            "\xe0\xe8\xe9\xf9\xec\xf2",
            "\xc3\xa0\xc3\xa8\xc3\xa9\xc3\xb9\xc3\xac\xc3\xb2"
        )

    def test_cp1251(self):
        self._test_encoding("cp1251",
            "\xc1\xc2\xc3\xc4\xc5\xc6\xc7\xc8",
            "\xd0\x91\xd0\x92\xd0\x93\xd0\x94\xd0\x95\xd0\x96\xd0\x97\xd0\x98"
        )

    def test_unicode_filename(self):
        #self.requireFeature(UnicodeFilename)

        #wt = self.make_branch_and_tree('.')
        #filename = self.alpha + '.txt'
        #self.build_tree_contents([ (filename, text_raw) ])
        #wt.add([filename ])
        #wt.commit("commit")

        #log = self.do_test("?cmd=content;rev=1;path="+filename)

        #filename_utf8

        #self.assertContains(log, filename_utf8)


        pass

    def test_unicode_commit_comment(self):

        self.encoding = 'utf8'

        wt = self.make_branch_and_tree('.')
        self.build_tree_contents([ ('data.txt', '123') ])
        wt.add(["data.txt" ])
        wt.commit(self.alpha)

        log = self.do_test("?cmd=content;rev=1;path=data.txt")

        self.assertContains(log, self.alpha.encode('utf8'))

    def test_unicode_author(self):

        def obfuscate(text):
            return ''.join(['&#%d;' % ord(c) for c in text])
        
        self.encoding = 'utf8'

        wt = self.make_branch_and_tree('.')
        self.build_tree_contents([ ('data.txt', '123') ])
        wt.add(["data.txt" ])
        wt.commit(u'commit-'+self.omega, committer='author'+self.alpha)

        log = self.do_test("?cmd=content;rev=1;path=data.txt")

        self.assertContains(log, obfuscate(self.alpha))


class TestTemplater(TestCaseInTempDir):

    def test_simple(self):
        map1 = {"simple": "simple"}

        t = templater(maptext=map1)
        s = StringIO.StringIO()

        write2(s, t("simple"))

        self.assertEqual(s.getvalue(), "simple")

    def test_with_static_text(self):
        map1 = {"medium": "medium: insert=#insert#"}

        t = templater(maptext=map1)
        s = StringIO.StringIO()

        write2(s, t("medium", insert="foo"))

        self.assertEqual(s.getvalue(), "medium: insert=foo")

    def test_with_function(self):
        map1 = {"medium": "medium: insert=#insert#"}

        def f(**args):
            return "foo"

        t = templater(maptext=map1)
        s = StringIO.StringIO()

        write2(s, t("medium", insert=f))

        self.assertEqual(s.getvalue(), "medium: insert=foo")

    def test_with_iterator(self):
        map1 = {"medium": "medium: insert=#insert#"}

        def f(**args):
            for i in range(2):
                yield "foo"

        t = templater(maptext=map1)
        s = StringIO.StringIO()

        write2(s, t("medium", insert=f))

        self.assertEqual(s.getvalue(), "medium: insert=foofoo")

    def test_with_nested_template(self):
        map1 = {"medium": "medium: insert=#insert#"}

        def f(**args):
            for i in range(2):
                yield "foo"

        t = templater(maptext=map1)
        s = StringIO.StringIO()

        t1 = t("medium", insert=f)
        t2 = t("medium", insert=t1)
        write2(s, t2)
        self.assertEqual(s.getvalue(), "medium: insert=medium: insert=foofoo")

    def test_with_function_and_filter(self):
        map1 = {"medium": "medium: insert=#insert|filter#"}

        def f(**args):
            return "foo"

        def f_filter(a):
            return "filtered(%s)"%a

        map_filter = {"filter": f_filter}

        t = templater(maptext=map1, filters=map_filter)
        s = StringIO.StringIO()

        write2(s, t("medium", insert=f))

        self.assertEqual(s.getvalue(), "medium: insert=filtered(foo)")

    def test_with_iterator_and_filter(self):
        map1 = {"medium": "medium: insert=#insert|filter#"}

        def f(**args):
            for i in range(2):
                yield "foo"

        def f_filter(a):
            return "filtered(%s)"%a

        map_filter = {"filter": f_filter}

        t = templater(maptext=map1, filters=map_filter)
        s = StringIO.StringIO()

        write2(s, t("medium", insert=f))

        self.assertEqual(s.getvalue(), "medium: insert=filtered(foo)filtered(foo)")

    def test_with_iterator_of_template_and_filter(self):
        map1 = {
            "filtered": "medium: insert=#insert|filter#",
            "simple":   "txt: insert=>#here#<",
        }

        def f_filter(a):
            return "filtered(%s)"%a

        map_filter = {"filter": f_filter}

        t = templater(maptext=map1, filters=map_filter)

        def f(**args):
            t1 = templater(maptext=map1, filters=map_filter)
            for i in range(2):
                yield t1("simple", here="foo")

        t = templater(maptext=map1, filters=map_filter)
        s = StringIO.StringIO()

        write2(s, t("filtered", insert=f))

        self.assertEqual(s.getvalue(),
          "medium: insert=filtered(txt: insert=>foo<)filtered(txt: insert=>foo<)")

    def test_with_function_and_multiple_filter(self):
        map1 = {"medium": "medium: insert=#insert|filter|filter#"}

        def f(**args):
            return "foo"

        def f_filter(a):
            return "filtered(%s)"%a

        map_filter = {"filter": f_filter}

        t = templater(maptext=map1, filters=map_filter)
        s = StringIO.StringIO()

        write2(s, t("medium", insert=f))

        self.assertEqual(s.getvalue(), "medium: insert=filtered(filtered(foo))")

    def test_with_iterator_and_filter_and_list_filter(self):
        map1 = {"medium": "medium: insert=#insert|filter||filter#"}

        def f(**args):
            for i in range(2):
                yield "foo"

        def f_filter(a):
            return "filtered(%s)"%a

        map_filter = {"filter": f_filter}

        t = templater(maptext=map1, filters=map_filter)
        s = StringIO.StringIO()

        write2(s, t("medium", insert=f))

        self.assertEqual(s.getvalue(), "medium: insert=filtered(filtered(foo)filtered(foo))")

    def test_with_iterator_and_filter_and_list_filter(self):
        map1 = {"medium": "medium: insert=#insert||filter#"}

        def f(**args):
            for i in range(2):
                yield "foo"

        def f_filter(a):
            return "filtered(%s)"%a

        map_filter = {"filter": f_filter}

        t = templater(maptext=map1, filters=map_filter)
        s = StringIO.StringIO()

        write2(s, t("medium", insert=f))

        self.assertEqual(s.getvalue(), "medium: insert=filtered(foofoo)")

    def test_with_iterator_and_multiple_filter_and_multiple_list_filter(self):
        map1 = {"medium": "medium: insert=#insert|filter|filter||filter|filter#"}

        def f(**args):
            for i in range(2):
                yield "foo"

        def f_filter(a):
            return "filtered(%s)"%a

        map_filter = {"filter": f_filter}

        t = templater(maptext=map1, filters=map_filter)
        s = StringIO.StringIO()

        write2(s, t("medium", insert=f))

        self.assertEqual(s.getvalue(), "medium: insert=filtered(filtered(filtered(filtered(foo))filtered(filtered(foo))))")

    def test_templater_with_a_loop(self):
        map1 = {"tmpl": "tmpl = #line#" }

        def f():
            yield "--> #line# <--"

        t = templater(maptext=map1)
        s = StringIO.StringIO()

        write2(s, t("tmpl", line=f))
        self.assertEqual(s.getvalue(), "tmpl = --> #line# <--")


